from time import sleep
from typing import BinaryIO, Union, Optional, IO, Callable
from urllib3 import HTTPResponse
from rkd.api.inputoutput import IO as RKDIO
from .exception import BufferingError

BUFFER_CALLABLE_DEF = Callable[[Optional[int]], bytes]


class StreamableBuffer(object):
    _read_callback: BUFFER_CALLABLE_DEF
    _close: callable
    _has_exited_with_failure: callable
    _description: str
    _buffer: Union[BinaryIO, Optional[IO[bytes]], HTTPResponse]
    _in_buffer: Optional[BUFFER_CALLABLE_DEF]
    _parent: Optional['StreamableBuffer']

    # pre-validation of the buffer on read()
    _pre_validation_taken_place: bool
    _pre_validation_sleep: int

    _io: RKDIO

    # stats
    _stat_total_read: int

    def __init__(self, io: RKDIO, read_callback: BUFFER_CALLABLE_DEF,
                 close_callback: callable,
                 eof_callback: callable,
                 is_success_callback: callable,
                 buffer: Union[BinaryIO, Optional[IO[bytes]], HTTPResponse],
                 has_exited_with_failure: callable = None,
                 description: str = '',
                 in_buffer: Optional[BUFFER_CALLABLE_DEF] = None,
                 parent: Optional['StreamableBuffer'] = None,
                 pre_validation_sleep: int = 5):

        self._io = io
        self._read_callback = read_callback
        self._close = close_callback
        self._is_eof = eof_callback
        self._is_success = is_success_callback
        self._has_exited_with_failure = has_exited_with_failure
        self._description = description
        self._buffer = buffer
        self._in_buffer = in_buffer
        self._parent = parent
        self._pre_validation_sleep = pre_validation_sleep

        self._pre_validation_taken_place = False
        self._stat_total_read = 0

    def get_buffer(self) -> Union[BinaryIO, Optional[IO[bytes]]]:
        return self._buffer

    def get_in_buffer(self) -> Optional[BUFFER_CALLABLE_DEF]:
        return self._in_buffer

    def read(self, size: int = 64 * 1024) -> bytes:
        """
        Read stream of given length
        At first read() call it performs stream validation to see if it didn't end prematurely

        :param size:
        :return:
        """

        buf = None

        if self._pre_validation_taken_place is False:
            self._io.debug('Validating stream against premature termination')

            self._pre_validation_taken_place = True

            buf = self._read_callback(size)
            # give the process / api request or other thing a little amount of time on first read
            sleep(self._pre_validation_sleep)

        if self.has_exited_with_failure():
            raise BufferingError.from_early_buffer_exit(self._description)

        self._stat_total_read += size
        self._io.debug('Read {size} (total={total}b, {total_mb}mb)'.format(size=size, total=self._stat_total_read,
                                                                           total_mb=self._stat_total_read/1024/1024))

        return self._read_callback(size) if buf is None else buf

    def read_all(self) -> bytes:
        # noinspection PyArgumentList
        return self._read_callback()

    def close(self):
        self._io.debug('Closing stream')

        return self._close()

    def eof(self) -> bool:
        return self._is_eof()

    def finished_with_success(self) -> bool:
        # fail, when parent fails
        if self._parent and not self._parent.finished_with_success():
            return False

        return self._is_success()

    def has_exited_with_failure(self) -> bool:
        """If stream comes from a process, then it will check if process not exited already with >= 1 exit code"""

        # not implemented
        if not self._has_exited_with_failure:
            return False

        # fail, when parent fails
        if self._parent and self._parent.has_exited_with_failure():
            return True

        return self._has_exited_with_failure()

    @staticmethod
    def from_file(path: str, io: IO) -> 'StreamableBuffer':
        handle = open(path, 'rb')

        return StreamableBuffer(
            io=io,
            read_callback=handle.read,
            close_callback=lambda: handle.close(),
            eof_callback=lambda: handle.closed,
            is_success_callback=lambda: handle.closed,
            description='File stream <{}>'.format(path),
            buffer=handle,
            pre_validation_sleep=0
        )

    def find_failure_cause(self) -> str:
        """Find a stream that broke the pipeline"""

        if self._parent and self._parent.has_exited_with_failure():
            return self._parent.find_failure_cause()

        if self.has_exited_with_failure():
            return self._description

        return ''



from rkd.api.inputoutput import IO
from rkd.api.testing import BasicTestingCase
from bahub.transports.temporarydocker import Transport


class TestTemporaryDockerTransport(BasicTestingCase):
    """
    Functional test - requires docker daemon and docker client tools
    """

    def test_captures_output(self):
        self.assertIn(b'PG_VERSION=13', self._create_example_transport().capture('env').strip())

    def test_buffers_output(self):
        buffer = self._create_example_transport().buffered_execute('ls -la /')
        buffer._pre_validation_sleep = 0  # hack to only speed up tests, normally it is a private method

        self.assertIn(b'tmp', buffer.read().strip())
        buffer.close()

    @staticmethod
    def _create_example_transport() -> Transport:
        return Transport(
            spec={
                'orig_container': 's3pb_db_postgres_1',
                'temp_container_image': 'postgres:13.1',
                'shell': '/bin/bash'
            },
            io=IO()
        )



import os
import sys
import gnupg
from typing import Union, List
from subprocess import Popen, PIPE
from rkd.api.inputoutput import IO
from .exception import CryptographyKeysAlreadyCreated
from .inputoutput import StreamableBuffer
from .model import Encryption


class EncryptionService(object):
    _io: IO

    def __init__(self, io: IO):
        self._io = io

    def create_decryption_stream(self, encryption: Encryption, stdin: Union[int, 'StreamableBuffer'] = PIPE) \
            -> StreamableBuffer:

        """Takes incoming encrypted buffer and returns a buffer that outputs a decrypted data"""

        gpg_command = [
            'gpg',
            '--homedir', encryption.get_home_dir(),
            '--decrypt',
            '--recipient', encryption.recipient(),
            '--armor',
            '--passphrase', encryption.get_passphrase(),
            '--batch',
            '--yes',
            '--pinentry-mode', 'loopback',
            '--verbose',
        ]

        return self._create_stream(encryption, gpg_command, stdin)

    def create_encryption_stream(self, encryption: Encryption, stdin: Union[int, 'StreamableBuffer'] = PIPE) \
            -> StreamableBuffer:

        gpg_command = [
            'gpg',
            '--homedir', encryption.get_home_dir(),
            '--encrypt',
            '--recipient', encryption.recipient(),
            '--armor'
        ]

        return self._create_stream(encryption, gpg_command, stdin)

    def _create_stream(self, encryption: Encryption, gpg_command: List[str],
                       stdin: Union[int, 'StreamableBuffer'] = PIPE) -> StreamableBuffer:

        self._io.debug('popen({})'.format(gpg_command))
        proc = Popen(gpg_command,
                     stdout=PIPE,
                     stdin=PIPE if stdin == PIPE else stdin.get_buffer(),
                     stderr=sys.stderr.fileno(),
                     close_fds=True
                     )

        return StreamableBuffer(
            io=self._io,
            read_callback=proc.stdout.read,
            close_callback=proc.terminate,
            eof_callback=lambda: proc.poll() is not None,
            is_success_callback=lambda: proc.poll() == 0,
            has_exited_with_failure=lambda: proc.poll() is not None and proc.poll() >= 1,
            description='Encryption stream <{}>'.format(str(gpg_command)),
            buffer=proc.stdout,
            in_buffer=proc.stdin,
            parent=stdin
        )

    def create_keys(self, enc: Encryption):
        """Generates GPG keys usable for backup encryption & decryption"""

        gpg = self._get_pgp(enc)

        for key in gpg.list_keys():
            if enc.get_userid() in str(key['uids']):
                raise CryptographyKeysAlreadyCreated.from_keys_already_created(enc.get_userid())

        result = gpg.gen_key(
            gpg.gen_key_input(
                key_type=enc.get_key_type(),
                key_length=enc.get_key_length(),
                name_real=enc.get_username(),
                name_comment="Generated by Bahub (https://github.com/riotkit-org/backup-repository)",
                name_email=enc.get_userid(),
                expire_date=0,
                passphrase=enc.get_passphrase()
            )
        )

        self._io.success_msg('Keys created, fingerprint: {fingerprint}'.format(fingerprint=result))

    def list_keys(self, enc: Encryption) -> List[dict]:
        """Lists GPG keys for given Encryption entry"""

        gpg = self._get_pgp(enc)

        return list(map(
            lambda key: {'email': key['uids'][0], 'fingerprint': key['fingerprint'], 'gpg_home': enc.get_home_dir()},
            gpg.list_keys()
        ))

    @staticmethod
    def _get_pgp(enc: Encryption) -> gnupg.GPG:
        if not os.path.isdir(enc.get_home_dir()):
            os.mkdir(enc.get_home_dir())

        gpg = gnupg.GPG(gnupghome=enc.get_home_dir())
        gpg.encoding = 'utf-8'

        return gpg




import os
from rkd import main as rkd_main, RiotKitDoApplication


def main():
    RiotKitDoApplication.load_environment()

    os.environ['RKD_PATH'] = os.path.dirname(os.path.realpath(__file__)) + '/internal:' + os.getenv('RKD_PATH', '')
    os.environ['RKD_DIST_NAME'] = 'bahub'
    os.environ['RKD_BIN'] = 'bahub'
    os.environ['RKD_UI'] = 'true'

    rkd_main()


if __name__ == '__main__':
    main()



import os
from tempfile import TemporaryDirectory
from rkd.api.inputoutput import IO
from rkd.api.testing import BasicTestingCase
from bahub.encryption import EncryptionService
from bahub.inputoutput import StreamableBuffer
from bahub.model import Encryption


class TestEncryptionService(BasicTestingCase):
    """
    Functional test - requires gpg binary to be present
    """

    def test_keys_creation_and_listing(self):
        """
        Functional test that checks if GPG keys are correctly created
        Basic successful case
        """

        io = IO()
        service = EncryptionService(io)

        with TemporaryDirectory() as tempdir:
            encryption = self._create_test_encryption(tempdir)

            # create key & list all keys
            service.create_keys(encryption)
            keys_list = service.list_keys(encryption)

            # only one key should be created
            self.assertEqual(
                'Mikhail Bakunin (Generated by Bahub (https://github.com/riotkit-org/backup-repository)) ' +
                '<bakunin@riotkit.local>', keys_list[0]['email'])

            self.assertIsNotNone(keys_list[0]['fingerprint'])
            self.assertEqual(tempdir, keys_list[0]['gpg_home'])
            self.assertEqual(1, len(keys_list))

    def test_encryption_and_decryption(self):
        """
        Generates a key, then encrypts and decrypts a file, compares decrypted file with original before encryption
        """

        io = IO()
        service = EncryptionService(io)

        with TemporaryDirectory() as tempdir:
            encryption = self._create_test_encryption(tempdir)

            try:
                plaintext = b'''
                    Common to all Anarchists is the desire to free society of all 
                    political and social coercive institutions which stand in the 
                    way of development of a free humanity.
                '''

                # prepare a test file for encryption
                with open(tempdir + '/plaintext', 'wb') as f:
                    f.write(plaintext)

                # 1. Create keys
                service.create_keys(encryption)

                # 2. Encrypt a file
                r, w = os.pipe()
                os.write(w, plaintext)
                os.close(w)

                plain_buffer_streamable = StreamableBuffer.from_file(tempdir + '/plaintext', io=io)

                encrypted_buffer = service.create_encryption_stream(encryption, stdin=plain_buffer_streamable)
                encrypted_text = encrypted_buffer.read(1024 * 1024 * 64)

                # 2.2. Verify contents - must contain GPG header and footer
                with self.subTest('Verify encrypted text header and footer'):
                    self.assertTrue(encrypted_text.startswith(b'-----BEGIN PGP MESSAGE-----'))
                    self.assertTrue(encrypted_text.endswith(b'-----END PGP MESSAGE-----\n'))

                # 3. Decrypt back a file
                with self.subTest('Verify decryption gives same result as original input before encryption'):
                    # we need to store the encrypted text first to open a normal buffer from file
                    with open(tempdir + '/encrypted.gpg', 'wb') as f:
                        f.write(encrypted_text)

                    decryption_buffer = service.create_decryption_stream(
                        encryption,
                        stdin=StreamableBuffer.from_file(tempdir + '/encrypted.gpg', io=io)
                    )

                    # 3.1. Compare original text with decrypted text
                    self.assertEqual(
                        plaintext,
                        decryption_buffer.read(1024 * 1024 * 64)
                    )

            finally:
                # clean up
                encrypted_buffer.close()
                plain_buffer_streamable.close()
                os.close(r)
                decryption_buffer.close()

    @staticmethod
    def _create_test_encryption(tempdir: str):
        return Encryption(
            name='Bakunin-Strong',
            passphrase='To revolt is a natural tendency of life. ' +
                       'Even a worm turns against the foot that crushes it.',
            username='Mikhail Bakunin',
            email='bakunin@riotkit.local',
            algorithm='aes256',
            gnupg_home_path=tempdir,
            key_length=1024,
            key_type='RSA'
        )



import json
import yaml
from argparse import ArgumentParser
from rkd.api.contract import ExecutionContext, TaskInterface
from ..importing import Importing
from ..transports import transports
from ..adapters import adapters
from .base import BaseTask


class BackupTypeSchemaPrintingTask(BaseTask):
    """Shows JSON schema of section 'spec' in backup of given type"""

    def get_name(self) -> str: return ':schema'
    def get_group_name(self) -> str: return ':help:backup'

    def configure_argparse(self, parser: ArgumentParser, with_definition: bool = False):
        super().configure_argparse(parser, with_definition)
        parser.add_argument('type', help='Task type (value of the field "type" in backup definitions)')

    def execute(self, context: ExecutionContext) -> bool:
        super().execute(context)

        task_type = context.get_arg('type')
        adapter, definition = Importing.import_adapter(task_type)

        self.io().outln(
            json.dumps(definition.get_specification_schema(), indent=4, sort_keys=True)
        )

        return True


class BackupTypeExampleTask(BaseTask):
    """Shows a example configuration for a given backup type. See :help:info for list of built-in tasks"""

    def get_name(self) -> str: return ':example'
    def get_group_name(self) -> str: return ':help:backup'

    def configure_argparse(self, parser: ArgumentParser, with_definition: bool = False):
        super().configure_argparse(parser, with_definition)
        parser.add_argument('type', help='Task type (value of the field "type" in backup definitions)')

    def execute(self, context: ExecutionContext) -> bool:
        super().execute(context)

        task_type = context.get_arg('type')
        adapter, definition = Importing.import_adapter(task_type)

        self.io().outln(
            yaml.dump({
                'backups': {
                    'example': definition.get_example_configuration()
                }
            }, indent=4)
        )

        return True


class TransportTypeTask(BaseTask):
    """Shows example transport configuration for given transport. See :help:info for list of built-in transports"""

    def get_name(self) -> str: return ':example'
    def get_group_name(self) -> str: return ':help:transport'

    def configure_argparse(self, parser: ArgumentParser, with_definition: bool = False):
        super().configure_argparse(parser, with_definition)
        parser.add_argument('type', help='Transport type (value of the field "type" in transport list)')

    def execute(self, context: ExecutionContext) -> bool:
        super().execute(context)

        transport_name = context.get_arg('type')
        transport = Importing.import_transport(transport_name)

        self.io().outln(
            yaml.dump({
                'transports': {
                    'example': transport.get_example_configuration()
                }
            }, indent=4)
        )

        return True


class InfoTask(TaskInterface):
    """Lists all built-in backup types and transports"""

    def get_name(self) -> str: return ':info'
    def get_group_name(self) -> str: return ':help'

    def execute(self, context: ExecutionContext):

        self.io().print_line()
        self.io().outln('Standard built-in transports:')

        for transport in transports():
            self.io().outln('- {}'.format(transport.__module__))

        self.io().print_line()
        self.io().outln('Standard built-in backup types:')

        for adapter in adapters():
            self.io().outln('- {}'.format(adapter.__module__))

        self.io().print_line()
        self.io().info_msg('*The list includes ONLY officially distributed adapters and transports. '
                           'You may be able to install additional adapters and transports via PyPI')

        self.io().info_msg('Check also help for transports: `bahub :help:transport:example bahub.transports.sh`')
        self.io().info_msg('Check also help for backup types: `bahub :help:backup:example bahub.adapters.filesystem`')

        return True

    def configure_argparse(self, parser: ArgumentParser):
        pass




from abc import ABC as AbstractClass, abstractmethod


class Response(AbstractClass):
    @abstractmethod
    def to_status_message(self) -> str:
        """Formats the response as user-readable to show user as a message"""

        pass


class VersionUploadedResponse(Response):
    version: str
    fileid: int
    filename: str

    def __init__(self, version: str, fileid: int, filename: str):
        self.version = version
        self.fileid = fileid
        self.filename = filename

    def to_status_message(self) -> str:
        return 'Uploaded version "{version}" of "{filename}"'.format(
            version=self.version,
            filename=self.filename
        )



"""
Temporary Docker Transport (bahub.transports.temporarydocker)
=============================================================

Places a temporary container to copy data from other container while the second container will be shutted down for
maintenance time
"""

from docker.errors import NotFound
from rkd.api.inputoutput import IO

from .docker import Transport as RegularDockerTransport
from ..exception import DockerContainerError


class Transport(RegularDockerTransport):
    """
    Behaves exactly the same as regular Docker Transport, but on a temporary container
    Even the methods are used from parent Docker Transport - and the `self._container_name` is replaced
    with a temporary container name
    """

    _original_container: str
    _temp_container_id: str

    def __init__(self, spec: dict, io: IO):
        super().__init__(spec, io)

        self._shell = spec.get('shell', '/bin/bash')
        self._container_name = spec.get('orig_container')

    @staticmethod
    def get_specification_schema() -> dict:
        return {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "required": ["temp_container_image", "orig_container"],
            "properties": {
                "orig_container": {
                    "type": "string",
                    "example": "bahub_adapter_integrations_db_mysql_1"
                },
                "temp_container_image": {
                    "type": "string",
                    "example": "mariadb:10.3"
                },
                "shell": {
                    "type": "string",
                    "example": "/bin/bash",
                    "default": "/bin/bash"
                }
            }
        }

    def __enter__(self) -> 'Transport':
        """
        Spawns a temporary container
        :return:
        """

        # stop the original container at first
        self._original_container = self._spec.get('orig_container')
        self._io.debug('Stopping original container {}'.format(self._original_container))
        self._client.stop(self._original_container)

        self._io.debug('Creating a temporary container...')
        host_config = self._client.create_host_config(volumes_from=[self._original_container])

        try:
            info = self._client.create_container(
                image=self._spec.get('temp_container_image'),
                entrypoint=['sleep'],
                command=[str(86400 * 5)],
                host_config=host_config
            )

            self._io.debug('Temporary container was spawned')
            self._temp_container_id = info['Id']
            self._container_name = self._temp_container_id

            self._client.start(self._container_name)
            self._io.debug('Temporary container was started')
        except:
            self._client.start(self._original_container)
            raise

        return self

    def __exit__(self, exc_type, exc_val, exc_t) -> None:
        try:
            self._io.debug('Killing temporary container')
            self._client.kill(self._container_name)
        finally:
            self._io.debug('Removing temporary container')
            self._client.remove_container(self._container_name, force=True)

            self._io.debug('Bringing back the orginal container')
            self._client.start(self._original_container)

    def _assert_container_is_fine(self):
        try:
            self._client.inspect_container(self._container_name)
        except NotFound:
            raise DockerContainerError.from_container_not_found(self._container_name)




from rkd.api.inputoutput import IO
from rkd.api.testing import BasicTestingCase
from bahub.transports.sh import Transport


class TestShellTransport(BasicTestingCase):
    """
    Functional test - requires /bin/bash
    """

    def test_captures_output(self):
        self.assertIn(b'usr', self._create_example_transport().capture('ls -la /').strip())

    def test_buffers_output(self):
        buffer = self._create_example_transport().buffered_execute('ls -la /')
        buffer._pre_validation_sleep = 0  # hack to only speed up tests, normally it is a private method

        self.assertIn(b'usr', buffer.read().strip())
        buffer.close()

    @staticmethod
    def _create_example_transport() -> Transport:
        return Transport(
            spec={'shell': '/bin/bash'},
            io=IO()
        )



"""
Backup Adapters
===============

Provides a dedicated backup handlers. Each backup handler is specialized in its own domain.

Example of handlers:
  - MySQL database handler
  - Filesystem handler
  - Redis handler
  - Remote Mail Handler
"""

from .filesystem import Adapter as FSAdapter
from .mysql import Adapter as MySQLAdapter
from .postgres_dump import Adapter as PostgresDumpAdapter


def adapters() -> list:
    return [FSAdapter, MySQLAdapter, PostgresDumpAdapter]



from rkd.api.testing import BasicTestingCase
from bahub.security import create_sensitive_data_stripping_filter


class TestSecurity(BasicTestingCase):
    def test_create_sensitive_data_stripping_filter_produces_method(self):
        """
        Test that method still is a factory method
        :return:
        """

        callable_method = create_sensitive_data_stripping_filter(['some', 'awesome'])

        self.assertEqual('function', type(callable_method).__name__)

    def test_create_sensitive_data_stripping_filter_replaces_selected_words(self):
        """
        Check that multiple occurrences are replaced
        :return:
        """

        words = ['PiS', 'Konfederacja', 'all', 'Trump', 'parties', 'fck']

        self.assertEqual(
            '******** ********, ******** and ********, and ******** ********. ******** them ********!',
            create_sensitive_data_stripping_filter(words)('fck PiS, Trump and Konfederacja, and all parties.' +
                                                          ' fck them all!')
        )



from argparse import ArgumentParser
from rkd.api.contract import ExecutionContext
from .base import BaseTask
from ..adapters.base import AdapterInterface
from ..exception import BackupProcessError
from ..inputoutput import StreamableBuffer
from ..transports.sh import Transport as LocalShellTransport


class RestoreTask(BaseTask):
    """Restores data from previously uploaded backup
    """

    def get_name(self) -> str:
        return ':restore'

    def get_group_name(self) -> str:
        return ':backup'

    def configure_argparse(self, parser: ArgumentParser, with_definition: bool = True):
        super().configure_argparse(parser, with_definition=with_definition)
        parser.add_argument('--path', required=False,
                            help='File path from the disk. Defaults to empty and downloading from server')
        parser.add_argument('--version', default='latest', help='Version number. Defaults to "latest"')
        parser.add_argument('--download', '-d', required=False, help='Set target path as argument to download only')

    def execute(self, context: ExecutionContext) -> bool:
        if not super().execute(context):
            return False

        definition_name = context.get_arg('definition')
        definition = self.retrieve_definition(context)
        version = context.get_arg('--version')
        download_to = context.get_arg('--download')
        backup_adapter: AdapterInterface = self.config.get_adapter(definition_name)()

        if not definition:
            return False

        #
        # There are 3 streams piping one->two->three
        #
        self.io().info('Opening a backup buffer from server')
        remote_file_buffer = self.api.read_backup(
            collection_id=definition.get_collection_id(),
            access=definition.access(),
            version=version
        )

        self.io().info('Creating a decryption buffer')
        enc_buffer = self.encryption_service.create_decryption_stream(definition.encryption(),
                                                                      stdin=remote_file_buffer)

        try:
            self.io().info('Spawning the backup adapter to start "restore" procedure specific to this backup type')
            self.io().info('Processing combined pipeline of buffers...')
            self.notifier.starting_backup_restore(definition)

            # only download into a file
            if download_to:
                self._download_only(download_to, enc_buffer)
            # restore
            else:
                with definition.transport():
                    backup_adapter.restore(definition, enc_buffer, self._io)
                    self.notifier.backup_was_restored(definition)

        except BackupProcessError as e:
            additional_info = definition.transport().get_failure_details()

            if additional_info:
                self.io().error_msg(additional_info)

            self.io().error_msg(e)
            self.notifier.failed_to_restore_backup(definition, str(e))

            return False

        self.io().info_msg('Backup {} was restored to {}'.format(definition, version))

        return True

    def _download_only(self, target_path: str, enc_buffer: StreamableBuffer):
        """Downloads a file into a target path. Useful for data inspection"""

        self.io().info('Downloading a file to "{}"'.format(target_path))

        sh = LocalShellTransport({}, self._io)
        buffer = sh.buffered_execute('cat - > "{}"'.format(target_path), stdin=enc_buffer)
        buffer.read_all()



import importlib
from typing import Tuple, Type
from .adapters.base import AdapterInterface
from .exception import ParsingException
from .model import BackupDefinition
from .transports.base import TransportInterface


class Importing(object):
    @staticmethod
    def import_adapter(import_str: str) -> Tuple[Type[AdapterInterface], Type[BackupDefinition]]:
        """
        Imports an Adapter - service + model (Based on rkd.api.parsing.SyntaxParsing)

        :raises ParsingException
        :return:
        """

        try:
            module = importlib.import_module(import_str)
        except ImportError as e:
            raise ParsingException.from_import_error(import_str, e)

        if "Adapter" not in dir(module):
            raise ParsingException.from_class_not_found_in_module_error(import_str, 'Adapter')

        if "Definition" not in dir(module):
            raise ParsingException.from_class_not_found_in_module_error(import_str, 'Definition')

        return module.Adapter, module.Definition

    @staticmethod
    def import_transport(import_str: str) -> Type[TransportInterface]:
        try:
            module = importlib.import_module(import_str)
        except ImportError as e:
            raise ParsingException.from_import_error(import_str, e)

        if "Transport" not in dir(module):
            raise ParsingException.from_class_not_found_in_module_error(import_str, 'Transport')

        return module.Transport



"""
Shell Transport
===============

Executes a command in the shell
"""
import os
import sys
from subprocess import check_call, check_output, Popen, PIPE
from typing import List, Union, Optional
from time import sleep
from ..inputoutput import StreamableBuffer
from .base import TransportInterface


class Transport(TransportInterface):
    """
    Local shell transport
    =====================

    Executes commands in a local shell
    """

    @staticmethod
    def get_specification_schema() -> dict:
        return {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "required": [],
            "properties": {

            }
        }

    def execute(self, command: Union[str, List[str]]):
        check_call(command, shell=type(command) == str)

    def capture(self, command: Union[str, List[str]]) -> bytes:
        return check_output(command, shell=type(command) == str)

    def buffered_execute(self, command: Union[str, List[str]],
                         stdin: Optional[StreamableBuffer] = None,
                         env: dict = None) -> StreamableBuffer:

        proc_env = dict(os.environ)

        if env:
            proc_env.update(env)

        self.io().debug('buffered_execute({command})'.format(command=command))
        proc = Popen(command, shell=type(command) == str,
                     stdout=PIPE,
                     stderr=sys.stderr.fileno(),
                     stdin=stdin.get_buffer() if stdin else PIPE,
                     env=env)

        def close_stream():
            proc.stdout.close()
            proc.terminate()
            sleep(1)

        return StreamableBuffer(
            io=self._io,
            read_callback=proc.stdout.read,
            close_callback=close_stream,
            eof_callback=lambda: proc.poll() is not None,
            is_success_callback=lambda: proc.poll() == 0,
            has_exited_with_failure=lambda: proc.poll() is not None and proc.poll() >= 1,
            description='Local Shell (SH) Transport stream <{}>'.format(command),
            buffer=proc.stdout,
            parent=stdin
        )




class ApplicationException(Exception):
    pass


class TransportException(ApplicationException):
    pass


class DockerContainerError(TransportException):
    @staticmethod
    def from_container_not_found(container_id: str) -> 'DockerContainerError':
        return DockerContainerError('Container "{}" is not created'.format(container_id))

    @staticmethod
    def from_container_not_running(container_id: str, status: str) -> 'DockerContainerError':
        return DockerContainerError('Container "{}" is not running but actually {}'.format(container_id, status))


class ConfigurationFactoryException(ApplicationException):
    pass


class ConfigurationError(ApplicationException):
    pass


class SpecificationError(ConfigurationError):
    pass


class ApiException(ApplicationException):
    pass


class HttpException(ApiException):
    pass


class InvalidResponseException(HttpException):
    _json = None
    _exit_code = 0

    def __init__(self, msg, _json: dict, _exit_code: int):
        super().__init__(msg)
        self._json = _json
        self._exit_code = _exit_code

    def get_json(self) -> dict:
        return self._json

    def get_error(self):
        return self._json['error'] if 'error' in self._json else str(self._json)


class ParsingException(ApplicationException):
    """Errors related to parsing YAML/Python syntax"""

    @classmethod
    def from_import_error(cls, import_str: str, error: Exception) -> 'ParsingException':
        return cls(
            'Import "%s" is invalid - error: %s' % (
                import_str, str(error)
            )
        )

    @classmethod
    def from_class_not_found_in_module_error(cls, import_str: str, class_name: str) -> 'ParsingException':
        return cls(
            'Import "%s" is invalid. Class or method "%s" not found in module' % (
                import_str, class_name
            )
        )


class BufferingError(ApplicationException):
    @staticmethod
    def from_early_buffer_exit(stream_description: str) -> 'BufferingError':
        return BufferingError('Buffering of stream "{}" ended earlier with error'.format(stream_description))


class CryptographyConfigurationError(ApplicationException):
    pass


class CryptographyKeysAlreadyCreated(CryptographyConfigurationError):
    @staticmethod
    def from_keys_already_created(user_id: str) -> 'CryptographyConfigurationError':
        return CryptographyKeysAlreadyCreated('Cryptography keys for "{uid}" already created, skipping creation'
                                              .format(uid=user_id))


class BackupProcessError(ApplicationException):
    """Errors related to backup/restore streaming"""


class BackupRestoreError(BackupProcessError):
    """When we cannot restore a backup - at least one stage failed in the process"""

    @staticmethod
    def from_generic_restore_failure(stream) -> 'BackupRestoreError':
        return BackupRestoreError("Restore error. Process exited with a failure - read logs above.\nFailure cause: {}"
                                  .format(stream.find_failure_cause()))



"""
Security
========

Provides security helpers
"""
from typing import List, Union


def create_sensitive_data_stripping_filter(words: List[str]) -> callable:
    """
    Factory method that creates a filtering method used to remove secrets from console output
    :param words: List of sensitive words to replace with asterisks
    :return:
    """

    def sensitive_word_filter(text, origin: Union[str, bytes] = '') -> Union[str, bytes]:
        for word in words:
            if word in text:
                text = text.replace(word, '********')

        return text

    return sensitive_word_filter



from rkd.api.testing import BasicTestingCase
from bahub.schema import create_example_from_attributes


class TestSchema(BasicTestingCase):
    def test_create_example_from_attributes(self):
        """
        Given there are JSON schema object attributes
        Need: create an example usage basing on the schema

        :return:
        """

        example = {
            'field1': {
                'type': 'string'
            },
            'field2': {
                'type': 'string',
                'example': 'login123'
            },
            'field3': {
                'type': 'object',
                'properties': {
                    'subfield3.1': {
                        'type': 'string',
                        'example': 'Rojava'
                    }
                }
            }
        }

        self.assertEqual(
            {
                'field1': 'string',
                'field2': 'login123',
                'field3': {
                    'subfield3.1': 'Rojava'
                }
            },
            create_example_from_attributes(example)
        )



import os
from abc import ABC
from argparse import ArgumentParser
from typing import Dict, Union, Optional
from rkd.api.contract import TaskInterface, ExecutionContext, ArgumentEnv
from rkd.yaml_parser import YamlFileLoader
from rkd.exception import YAMLFileValidationError
from ..api import BackupRepository
from ..configurationfactory import ConfigurationFactory
from ..encryption import EncryptionService
from ..model import BackupDefinition
from ..notifier import MultiplexedNotifiers, NotifierInterface
from ..security import create_sensitive_data_stripping_filter


class BaseTask(TaskInterface, ABC):
    config: ConfigurationFactory
    api: BackupRepository
    notifier: Union[MultiplexedNotifiers, NotifierInterface]
    encryption_service: EncryptionService

    def get_declared_envs(self) -> Dict[str, Union[str, ArgumentEnv]]:
        return {
            'CONFIG': ArgumentEnv('CONFIG', '--config', os.path.expanduser('~/.bahub.yaml')),
            'DEBUG': ArgumentEnv('DEBUG', '--debug', '')
        }

    def execute(self, context: ExecutionContext) -> bool:
        try:
            self.config = ConfigurationFactory(
                configuration_path=context.get_arg_or_env('--config'),
                debug=bool(context.get_arg_or_env('--debug')),
                parser=YamlFileLoader([]),
                io=self._io
            )

        except YAMLFileValidationError as e:
            self.io().error('Configuration file looks invalid, details: ' + str(e))
            return False

        if not context.get_arg('--show-secrets'):
            self._io.add_output_processor(create_sensitive_data_stripping_filter(self.config.get_all_sensitive_data()))

        self.api = BackupRepository(self._io)
        self.notifier = MultiplexedNotifiers(self.config.notifiers())
        self.encryption_service = EncryptionService(self._io)

        return True

    def configure_argparse(self, parser: ArgumentParser, with_definition: bool = True):
        # do not require as switch, allow to use env
        parser.add_argument('--config', '-c', default=os.path.expanduser('~/.bahub.yaml'), required=False)
        parser.add_argument('--debug', action='store_true')
        parser.add_argument('--show-secrets', action='store_true', help='Do not hide secrets in output')

        if with_definition:
            parser.add_argument('definition', help='Backup definition name from the configuration file')

    def retrieve_definition(self, context: ExecutionContext) -> Optional[BackupDefinition]:
        definition_name = context.get_arg('definition')
        definition = self.config.get_definition(definition_name)

        if not definition:
            self.io().error('Definition "{name}" is not valid'.format(name=definition_name))

        return definition



"""
Core Tasks
==========

Provides shell commands such as:
- preparing backups
- sending backups
- listing backups
- receiving backups

"""
from rkd.api.syntax import TaskDeclaration
from .prepare import BackupPreparationTask
from .uploader import UploaderTask
from .restore import RestoreTask
from .cryptsetup import CryptographyKeysSetupTask, ListCryptoKeys
from .docs import BackupTypeSchemaPrintingTask, BackupTypeExampleTask, TransportTypeTask, InfoTask


def imports():
    return [
        TaskDeclaration(UploaderTask()),
        TaskDeclaration(RestoreTask()),
        TaskDeclaration(BackupPreparationTask()),
        TaskDeclaration(CryptographyKeysSetupTask()),
        TaskDeclaration(ListCryptoKeys()),
        TaskDeclaration(BackupTypeSchemaPrintingTask()),
        TaskDeclaration(BackupTypeExampleTask()),
        TaskDeclaration(TransportTypeTask()),
        TaskDeclaration(InfoTask())
    ]




class ErrorHandlerInterface:
    """
    Interface to an external monitoring service
    """

    config = {}

    def __init__(self, config: dict):
        self._set_config(config)
        self._initialize()

    def _set_config(self, config: dict):
        pass

    def _initialize(self):
        pass

    def record_exception(self, exception: BaseException):
        pass


class ErrorHandlerService:
    _handlers = []  # type: list[ErrorHandlerInterface]

    def __init__(self, handlers: dict):
        for k, v in handlers.items():
            self._handlers.append(v)

    def record_exception(self, exception: BaseException):
        for handler in self._handlers:
            handler.record_exception(exception)


class SentryIO(ErrorHandlerInterface):
    _exception_callback = None

    def _set_config(self, config: dict):
        if 'url' not in config:
            raise Exception('"url" is required for SentryIO configuration')

        self.config = config

    def _initialize(self):
        try:
            from sentry_sdk import init as sentry_init
            from sentry_sdk import capture_exception as sentry_capture_exception
            self._exception_callback = sentry_capture_exception

            sentry_init(self.config['url'])

        except ImportError:
            raise Exception('Cannot initialize SentryIO integration, sentry probably not installed')

    def record_exception(self, exception: BaseException):
        self._exception_callback(exception)


class ErrorHandlerFactory:
    _SUPPORTED = {
        'sentry': SentryIO,
        'sentry.io': SentryIO
    }

    @staticmethod
    def create(type_name: str, config: dict) -> ErrorHandlerInterface:
        if type_name not in ErrorHandlerFactory._SUPPORTED:
            raise Exception('Error handler type "' + type_name + '" not supported')

        return ErrorHandlerFactory._SUPPORTED[type_name](config)




import os
import re
from typing import Tuple, Type, Union, Dict, List
from json import dumps as json_encode, loads as json_decode
from rkd.api.inputoutput import IO
from rkd.yaml_parser import YamlFileLoader
from .adapters.base import AdapterInterface
from .importing import Importing
from .model import Encryption
from .model import ServerAccess
from .model import BackupDefinition
from .errorhandler import ErrorHandlerFactory, ErrorHandlerInterface
from .notifier import NotifierInterface, NotifierFactory
from .exception import ConfigurationFactoryException, ConfigurationError, SpecificationError
from .transports.base import TransportInterface


class ConfigurationFactory(object):
    """ Constructs objects basing on the configuration file """

    _accesses: dict        # type: dict[ServerAccess]
    _encryption: dict      # type: dict[Encryption]
    _backups: dict         # type: dict[BackupDefinition]
    _error_handlers: dict  # type: dict[ErrorHandlerInterface]
    _notifiers: dict       # type: dict[NotifierInterface]
    _transports: dict      # type: dict[TransportInterface]
    _adapters: dict        # type: dict[Type[AdapterInterface]]
    _debug = False         # type: bool
    _config_dir: str
    _io: IO

    def __init__(self, configuration_path: str, debug: bool, parser: YamlFileLoader, io: IO):
        self._io = io
        self._debug = debug
        self._config_dir = os.path.abspath(os.path.dirname(configuration_path))
        self._parse(self._process_env_variables(parser.load_from_file(configuration_path, 'org.riotkit.bahub')))

    def _parse(self, config: dict):
        self._accesses = {}
        self._encryption = {}
        self._backups = {}
        self._error_handlers = {}
        self._notifiers = {}
        self._transports = {}
        self._adapters = {}

        self._parse_accesses(config['accesses'])
        self._parse_encryption(config['encryption'])
        self._parse_transports(config['transports'])
        self._parse_backups(config['backups'])
        self._parse_monitoring_error_handlers(config.get('error_handlers', {}))
        self._parse_notifiers(config.get('notifiers', {}))

    def _process_env_variables(self, content: Union[dict, list]) -> Union[dict, list]:
        content_as_str: str = json_encode(content)

        env_list = list(dict(os.environ).items())
        env_list.sort(key=lambda item: (-len(item[0]), item[0]))

        # add special variables:
        env_list.append(['CONFIG_DIR', self._config_dir])

        for env in env_list:
            content_as_str = content_as_str.replace('${' + env[0] + '}', env[1])

        invalid_vars = set(re.findall('\${([A-Z0-9a-z_]+)}', content_as_str))

        if len(invalid_vars) > 0:
            raise ConfigurationFactoryException(
                'Following environment variables are not resolved: ' + (', '.join(invalid_vars))
            )

        return json_decode(content_as_str)

    def _parse_accesses(self, config: dict):
        """Access tokens"""

        for key, values in config.items():
            with DefinitionFactoryErrorCatcher('accesses.' + key, self._debug):
                self._accesses[key] = ServerAccess.from_config(values)

    def _parse_encryption(self, config: dict):
        """Security - Encryption"""

        for enc_name, config in config.items():
            with DefinitionFactoryErrorCatcher('encryption.' + enc_name, self._debug):
                try:
                    self._encryption[enc_name] = Encryption.from_config(enc_name, config)

                except KeyError as config_key_name:
                    raise ConfigurationError('Encryption "%s" is missing "%s" configuration option' %
                                             (enc_name, config_key_name))

    def _parse_transports(self, config: dict) -> None:
        """Transports - eg. sh, docker"""

        for transport_name, config in config.items():
            with DefinitionFactoryErrorCatcher('transports.' + transport_name, self._debug):
                try:
                    transport: Union[Type[TransportInterface], TransportInterface.__init__] = \
                        Importing.import_transport(config['type'])

                    transport.validate_spec(config['spec'])

                    self._transports[transport_name] = transport(config['spec'], self._io)

                except KeyError as config_key_name:
                    raise ConfigurationError('Transport "%s" is missing "%s" configuration option' %
                                             (transport_name, config_key_name))

    def _parse_backups(self, config: dict):
        """Backups"""

        for backup_name, config in config.items():
            with DefinitionFactoryErrorCatcher('backups.' + backup_name, self._debug):

                # find related access and encryption
                if not config['meta']['access'] in self._accesses.keys():
                    raise ConfigurationError('Backup "%s" has incorrect "access" specified' % backup_name)

                config['meta']['access'] = self._accesses[config['meta']['access']]

                if "encryption" in config['meta']:
                    config['meta']['encryption'] = self._encryption[config['meta']['encryption']]

                config['meta']['transport'] = self._transports[config['meta']['transport']]

                adapter, definition = self._create_definition(
                    config['meta']['type'],
                    config,
                    backup_name
                )

                self._backups[backup_name] = definition
                self._adapters[backup_name] = adapter

    @staticmethod
    def _create_definition(def_type: str, config: dict,
                           name: str) -> Tuple[Type[AdapterInterface], BackupDefinition]:
        """Imports and creates a BackupDefinition basing on selected adapter that implements AdapterInterface"""

        adapter, definition = Importing.import_adapter(def_type)

        return adapter, adapter.create_definition(config, name)

    def _parse_monitoring_error_handlers(self, config: dict):
        """Error handlers integration"""

        for key, values in config.items():
            with DefinitionFactoryErrorCatcher('error_handlers.' + key, self._debug):

                if 'type' not in values:
                    raise ConfigurationFactoryException('Error handler must have a defined type')

                self._error_handlers[key] = ErrorHandlerFactory.create(values['type'], values)

    def _parse_notifiers(self, config: dict):
        """Notifiers"""

        sensitive_data = []

        for definition in self._backups.values():
            sensitive_data += definition.get_sensitive_information()

        for key, values in config.items():
            with DefinitionFactoryErrorCatcher('notifiers.' + key, self._debug):

                if 'type' not in values:
                    raise ConfigurationFactoryException('Notifier type needs to be specified')

                notifier = NotifierFactory.create(values['type'], values, self._io)
                self._notifiers[key] = notifier

                # make the notification to be stripped out of sensitive data such as passwords
                notifier.set_sensitive_data_to_strip_out(sensitive_data)

    def get_error_handlers(self):
        return self._error_handlers

    def notifiers(self):
        return self._notifiers

    def definitions(self) -> Dict[str, BackupDefinition]:
        return self._backups

    def get_definition(self, name: str) -> BackupDefinition:
        if name not in self._backups:
            raise ConfigurationFactoryException(
                'No such backup definition, maybe a typo? Please check the configuration file'
            )

        return self._backups[name]

    def get_adapter(self, definition_name: str) -> Type[AdapterInterface]:
        """
        Finds and returns an AdapterInterface type (not a class object)
        """

        if definition_name not in self._adapters:
            raise ConfigurationFactoryException(
                'No such backup definition, maybe a typo? Please check the configuration file'
            )

        return self._adapters[definition_name]

    def get_all_sensitive_data(self) -> List[str]:
        """
        Collects sensitive data like passwords from the configuration file
        :return:
        """

        sensitive_data = []

        for access in self._accesses:
            sensitive_data.append(self._accesses[access].get_token())

        for backup in self._backups:
            if self._backups[backup].encryption().get_passphrase():
                sensitive_data.append(self._backups[backup].encryption().get_passphrase())

            sensitive_data += self._backups[backup].get_sensitive_information()

        return sensitive_data


class DefinitionFactoryErrorCatcher:
    _key_name = ""
    _debug = False

    def __init__(self, config_key_name: str, _debug: bool):
        self._key_name = config_key_name
        self._debug = _debug

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            if self._debug:
                return

            if exc_type == SpecificationError or exc_type == ConfigurationError:
                raise ConfigurationFactoryException('Error parsing {}, details: {}'.format(self._key_name, exc_val))

            raise ConfigurationFactoryException(
                ' ERROR: There was a problem during parsing the configuration at section "' +
                self._key_name + '" in key ' + str(exc_val) + '. Possibly the key is missing. Details: ' + str(exc_type))




"""
Server API connector
====================

Notice: Only E2E tests coverage
"""
import certifi
import pycurl
from subprocess import Popen, PIPE
from io import BytesIO
from json import JSONDecodeError, loads as json_loads
from rkd.api.inputoutput import IO

from bahub.env import is_curl_debug_mode
from bahub.exception import InvalidResponseException, HttpException
from bahub.inputoutput import StreamableBuffer
from bahub.model import ServerAccess, VersionAttributes
from bahub.response import VersionUploadedResponse

SEND_BACKUPS_URL = '/api/stable/repository/collection/{collectionId}/versions'
RECEIVE_BACKUPS_URL = '/api/stable/repository/collection/{collectionId}/backup/{version}'


class BackupRepository(object):
    def __init__(self, io: IO):
        self.io = io

    def send_backup(self, collection_id: str, access: ServerAccess, attributes: VersionAttributes,
                    source: StreamableBuffer) -> VersionUploadedResponse:

        response_body_stream = BytesIO()
        url = access.build_url(
            endpoint=SEND_BACKUPS_URL.replace('{collectionId}', collection_id),
            attributes=attributes
        )

        curl = pycurl.Curl()
        curl.setopt(curl.URL, url)
        curl.setopt(curl.CAINFO, certifi.where())
        curl.setopt(curl.CUSTOMREQUEST, 'POST')
        curl.setopt(curl.UPLOAD, 1)
        curl.setopt(curl.READFUNCTION, source.read)
        curl.setopt(curl.WRITEFUNCTION, response_body_stream.write)
        curl.setopt(curl.HTTPHEADER, ["Authorization: Bearer " + access.get_token()])

        # debug
        if is_curl_debug_mode():
            curl.setopt(curl.VERBOSE, True)
            curl.setopt(pycurl.SSL_VERIFYPEER, 0)
            curl.setopt(pycurl.SSL_VERIFYHOST, 0)

        # explicit timeouts
        curl.setopt(curl.CONNECTTIMEOUT, 300)
        curl.setopt(curl.TIMEOUT, 3600*24)

        try:
            curl.perform()

        except pycurl.error as e:
            self.io.error('Curl error cause: {}'.format(e.__cause__))

            raise HttpException(
                'HTTP request error: ' + str(e) + '. Probably the application ' +
                'backup process exited or timed out unexpectedly. Read above messages for details')

        response_body = response_body_stream.getvalue().decode('utf-8')

        self.io.debug('Request: ' + str(url))
        self.io.debug('response_body=' + response_body)

        try:
            _json = json_loads(response_body)
        except JSONDecodeError:
            _json = {}

        response_code = curl.getinfo(pycurl.HTTP_CODE)

        self.io.debug('response_code={}'.format(response_code))

        if response_code >= 400:
            raise InvalidResponseException('HTTP error, code {}, body: {}'.format(response_code, response_body), _json,
                                           _json.get('error_code', 0))

        curl.close()

        return VersionUploadedResponse(
            version=_json['version']['version'],
            fileid=_json['version']['id'],
            filename=_json['version']['file']['filename']
        )

    def read_backup(self, collection_id: str, version: str, access: ServerAccess) -> StreamableBuffer:
        url = access.build_url(
            RECEIVE_BACKUPS_URL.format(collectionId=collection_id, version=version)
        )

        process = ['curl', '--silent', '-X', 'GET', '-H', 'Authorization: Bearer {}'.format(access.get_token()), url]

        self.io.debug('read_backup({})'.format(process))

        # cannot be done with requests or urllib3 due to raw IO stream handles required to copy streams
        # from process to process. Requests uses urllib3, which produces invalid streams that produces deadlocks
        proc = Popen(process, stdout=PIPE)

        return StreamableBuffer(
            io=self.io,
            read_callback=proc.stdout.read,
            close_callback=lambda: proc.stdout.close(),
            eof_callback=lambda: proc.poll() is not None,
            is_success_callback=lambda: proc.poll() == 0,
            has_exited_with_failure=lambda: proc.poll() != 0,
            description='API file read stream<{}, {}>'.format(collection_id, version),
            buffer=proc.stdout
        )



"""
Docker Transport
================

Executes commands inside a running docker container.

Note: Requires access to the docker daemon. Make sure your user is in a "docker" group (has access to the socket)
"""
from docker import Client as DockerClient
from docker.errors import NotFound
from typing import List, Union, Optional
from rkd.api.inputoutput import IO
from ..exception import DockerContainerError
from ..inputoutput import StreamableBuffer
from .sh import Transport as ShellTransport


class Transport(ShellTransport):
    """
    Docker Transport
    ================

    Extends a standard shell transport by `docker exec` calls with proper escaping

    Why is docker-py not used fully? Because it lacks support for stdin & stdout read-write at one time in exec_start()
    which makes it unusable in buffered_execute()
    """

    _container_name: str
    _shell: str
    _client: DockerClient

    def __init__(self, spec: dict, io: IO):
        super().__init__(spec, io)

        self._spec = spec
        self._io = io
        self._container_name = spec.get('container')
        self._shell = spec.get('shell', '/bin/sh')
        self._client = DockerClient.from_env()

    @staticmethod
    def get_specification_schema() -> dict:
        return {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "required": ["container"],
            "properties": {
                "container": {
                    "type": "string",
                    "example": "bahub_adapter_integrations_db_mysql_1"
                },
                "shell": {
                    "type": "string",
                    "example": "/bin/sh",
                    "default": "/bin/sh"
                }
            }
        }

    def execute(self, command: Union[str, List[str]]):
        self._assert_container_is_fine()

        super().execute(self._create_command(command))

    def capture(self, command: Union[str, List[str]]) -> bytes:
        self._assert_container_is_fine()

        return super().capture(self._create_command(command))

    def buffered_execute(self, command: Union[str, List[str]],
                         stdin: Optional[StreamableBuffer] = None,
                         env: dict = None) -> StreamableBuffer:

        self._assert_container_is_fine()

        return super().buffered_execute(
            command=self._create_command(command),
            stdin=stdin,
            env=env
        )

    def _create_command(self, cmd: Union[str, List[str]]) -> List[str]:
        """Creates a docker exec command, properly escaped"""

        prepared_inside_cmd = '' if type(cmd) == list else cmd

        if not prepared_inside_cmd:
            for param in cmd:
                if ' ' in param:
                    param = '"{}"'.format(param.replace('"', '\"'))

                prepared_inside_cmd += param + ' '

        return ['docker', 'exec', '-i', self._container_name, self._shell, '-c', prepared_inside_cmd]

    def get_failure_details(self) -> str:
        return 'Error occurred while trying to execute command inside docker container - {}, logs: {}'.format(
            self._container_name,
            self._try_to_collect_logs()
        )

    def _assert_container_is_fine(self):
        try:
            inspected = self._client.inspect_container(self._container_name)
        except NotFound:
            raise DockerContainerError.from_container_not_found(self._container_name)

        if not inspected['State']['Running']:
            raise DockerContainerError.from_container_not_running(self._container_name, inspected['State']['Status'])

    def _try_to_collect_logs(self) -> str:
        """
        Try to collect logs (if it is possible) from the container
        """

        try:
            return self._client.logs(tail=15, follow=False, stream=False)

        except Exception:
            return '-- No logs --'



from argparse import ArgumentParser
from rkd.api.contract import ExecutionContext
from .base import BaseTask
from ..encryption import EncryptionService
from ..exception import CryptographyKeysAlreadyCreated


class CryptographyKeysSetupTask(BaseTask):
    """Generates OpenGPG keys required for encryption.
    Takes Backup Definition as input, then generates keys for connected Access to that Backup Definition.

    Notice: Multiple Backup Definitions could share same Accesses"""

    def get_name(self) -> str: return ':generate-keys'
    def get_group_name(self) -> str: return ':crypto'

    def configure_argparse(self, parser: ArgumentParser, with_definition: bool = True):
        super().configure_argparse(parser, with_definition=with_definition)
        parser.add_argument('--ignore-errors', action='store_true',
                            help='Do not panic when the keys are already created')

    def execute(self, context: ExecutionContext) -> bool:
        if not super().execute(context):
            return False

        definition_name = context.get_arg('definition')
        definition = self.config.get_definition(definition_name)
        ignore_errors = bool(context.get_arg('--ignore-errors'))

        try:
            EncryptionService(self._io).create_keys(definition.encryption())

        except CryptographyKeysAlreadyCreated as e:
            self.io().error_msg(e) if not ignore_errors else self.io().info_msg(e)
            return ignore_errors

        return True


class ListCryptoKeys(BaseTask):
    """List all cryptographic keys"""

    def get_name(self) -> str: return ':list-keys'
    def get_group_name(self) -> str: return ':crypto'

    def configure_argparse(self, parser: ArgumentParser, with_definition: bool = False):
        super().configure_argparse(parser, with_definition=with_definition)
        parser.add_argument('--ignore-errors', action='store_true',
                            help='Do not panic when the keys are already created')

        parser.add_argument('--definition', dest='definition', required=False,
                            help='Backup definition name to limit the list by (optional)')

    def execute(self, context: ExecutionContext) -> bool:
        if not super().execute(context):
            return False

        definition_name = context.get_arg('--definition')

        definitions = [self.config.get_definition(definition_name)] if definition_name else self.config.definitions()\
            .values()
        body = []

        for definition in definitions:
            keys = EncryptionService(self._io).list_keys(definition.encryption())
            body += list(map(lambda key: [
                definition.name(),
                definition.encryption().name(),
                key['fingerprint'],
                key['email'],
                key['gpg_home']
            ], keys))

        self.io().outln(self.table(
            header=['Backup definition', 'Access name', 'Fingerprint', 'Details', 'GPG Directory'],
            body=body))

        return len(body) > 0



"""
Transport
=========

Abstraction that allows end-user to select if the backup should be done via local shell, ssh, docker,
kubernetes container etc.

Any transport can be implemented that will take a command, then process and properly output the result.
"""

from .docker import Transport as DockerTransport
from .sh import Transport as ShellTransport
from .temporarydocker import Transport as TemporaryDockerTransport


def transports():
    return [ShellTransport, DockerTransport, TemporaryDockerTransport]



from shutil import copyfileobj
from argparse import ArgumentParser
from rkd.api.contract import ExecutionContext
from .base import BaseTask
from ..adapters.base import AdapterInterface
from ..exception import BufferingError, InvalidResponseException
from ..inputoutput import StreamableBuffer
from ..model import BackupDefinition


class BackupPreparationTask(BaseTask):
    """Makes a backup and prints to stdout or into a file specified by --target parameter
    """

    def get_name(self) -> str:
        return ':make'

    def get_group_name(self) -> str:
        return ':backup'

    def configure_argparse(self, parser: ArgumentParser, with_definition: bool = True):
        super().configure_argparse(parser, with_definition=with_definition)
        parser.add_argument('--target', default='/dev/stdout', help='Target path where to output backup '
                                                                    '(if backup should be outputted)')
        parser.add_argument('--dont-send', action='store_true', help='Dont send backup to server, instead prepare and'
                                                                     'output to file or stdout as specified by '
                                                                     '--target')

    def execute(self, context: ExecutionContext) -> bool:
        if not super().execute(context):
            return False

        definition_name = context.get_arg('definition')
        target_path = context.get_arg('--target')
        dont_send = context.get_arg('--dont-send')
        definition = self.config.get_definition(definition_name)
        adapter: AdapterInterface = self.config.get_adapter(definition_name)()

        # begin a backup, get a buffered reader
        with definition.transport():
            self.notifier.starting_backup_creation(definition)

            backup_buffer = adapter.backup(definition)

            # buffers
            enc_buffer = self.encryption_service.create_encryption_stream(definition.encryption(),
                                                                          stdin=backup_buffer)

            if target_path and target_path != '/dev/stdout':
                save_status = self._save_backup(target_path=target_path, source=enc_buffer)
            elif not dont_send:
                save_status = self._send_backup(definition=definition, source=enc_buffer)
            else:
                return False

            backup_buffer.close()

            is_success = backup_buffer.finished_with_success() and enc_buffer.finished_with_success() and save_status

        if not is_success:
            self.io().error_msg('Backup process did not return success. Check previous messages for details')

        return is_success

    def _save_backup(self, target_path: str, source: StreamableBuffer):
        out = open(target_path, 'wb')

        # copy encrypted stream to destination
        try:
            copyfileobj(source, out)

        except BufferingError as e:
            out.close()
            self.io().error_msg('Backup process died unexpectedly at early buffering stage. '
                                'The errors from stderr should be above, if there were any')
            self.io().error_msg('Details: {}'.format(str(e)))

            return False

        out.close()
        return True

    def _send_backup(self, definition: BackupDefinition, source: StreamableBuffer):
        try:
            response = self.api.send_backup(
                collection_id=definition.get_collection_id(),
                access=definition.access(),
                attributes=definition.encryption().describe_as_attributes(),
                source=source
            )

            self.notifier.backup_was_uploaded(definition)

        except InvalidResponseException as response_exc:
            self.io().error(str(response_exc))
            self.io().error(response_exc.get_error())
            self.notifier.failed_to_upload_backup(definition, response_exc.get_error())

            return False

        if response:
            self.io().outln(response.to_status_message())
            return True



"""
JSON Schema helper methods
==========================
"""

from typing import Dict


def create_example_from_attributes(attributes: dict) -> Dict[str, str]:
    """
    Creates an example usage - part of YAML document basing on the schema

    :param attributes:
    :return:
    """

    as_dict = {}

    for key, attribute in attributes.items():
        as_dict[key] = attribute['example'] if 'example' in attribute else attribute['type']

        if 'properties' in attribute:
            as_dict[key] = create_example_from_attributes(attribute['properties'])

    return as_dict



import os
from urllib.parse import quote as url_quote
from json import dumps as json_dumps
from abc import ABC, abstractmethod
from jsonschema import validate, draft7_format_checker, ValidationError
from .exception import ConfigurationFactoryException, SpecificationError
from .transports.base import TransportInterface
from .schema import create_example_from_attributes


class VersionAttributes(object):
    """Key=Value storage per file"""

    _store: dict

    def __init__(self, attributes: dict):
        self._store = attributes

    def get(self, name: str) -> str:
        return self._store[name]

    def set(self, name: str, value: str):
        self._store[name] = value

    def get_crypto_initialization_vector(self) -> str:
        return self.get('iv')

    def set_crypto_initialization_vector(self, value: str):
        self.set('iv', value)

    @staticmethod
    def create_empty():
        return VersionAttributes({})

    def serialize_to_querystring_value(self) -> str:
        return url_quote(json_dumps(self._store))


class ServerAccess(object):
    """Backup Repository user access"""

    _url = ""    # type: str
    _token = ""  # type: str

    def __init__(self, url: str, token: str):
        self._url = url
        self._token = token

    @staticmethod
    def from_config(config: dict):
        return ServerAccess(config['url'], config['token'])

    def get_url(self):
        return self._url

    def get_token(self):
        return self._token

    def build_url(self, endpoint: str, attributes: VersionAttributes = None) -> str:

        url = self._url.rstrip('/') + '/' + endpoint.lstrip('/')
        qs = '?'

        if attributes:
            qs += '&kv=' + attributes.serialize_to_querystring_value()

        return url + qs.replace('?&', '?')


class Encryption(object):
    """ Cryptography support (using OpenSSL command implementation) """

    _name: str
    _passphrase: str
    _algorithm: str
    _gnupg_home_path: str
    _key_length: int
    _key_type: str
    _username: str
    _user_email: str

    SUPPORTED_ALGORITHMS = ['aes256']

    def __init__(self, name: str, passphrase: str, username: str, email: str, algorithm: str = 'aes256',
                 gnupg_home_path: str = '~/.bahub-gnupg', key_length: int = 2048,
                 key_type: str = 'RSA'):

        self._name = name
        self._passphrase = passphrase
        self._algorithm = algorithm
        self._gnupg_home_path = os.path.expanduser(gnupg_home_path)
        self._key_length = key_length
        self._key_type = key_type
        self._username = username
        self._user_email = email

        if algorithm not in self.SUPPORTED_ALGORITHMS:
            raise ConfigurationFactoryException('Crypto "' + algorithm + '" is not supported. Please use one of: ' +
                                                str(self.SUPPORTED_ALGORITHMS))

    @staticmethod
    def from_config(name: str, config: dict):
        is_encrypting = config.get('method', '') != ''

        if is_encrypting:
            for key in ['passphrase']:
                if key not in config:
                    raise KeyError(key)

        return Encryption(
            name=name,
            passphrase=config.get('passphrase', ''),
            algorithm=config['method'],
            username=config['username'],
            email=config['email'],
            gnupg_home_path=config.get('gnupg_home', '~/.bahub-gnupg'),
            key_length=int(config.get('key_length', 2048)),
            key_type=config.get('key_type', 'RSA')
        )

    def describe_as_attributes(self) -> VersionAttributes:
        return VersionAttributes({
            'recipient': self.recipient(),
            'algorithm': self._algorithm
        })

    def get_home_dir(self) -> str:
        return self._gnupg_home_path

    def get_key_length(self) -> int:
        return self._key_length

    def get_key_type(self) -> str:
        return self._key_type

    def get_username(self) -> str:
        return self._username

    def get_userid(self) -> str:
        return self._user_email

    def recipient(self):
        return self.get_userid()

    def get_passphrase(self) -> str:
        return self._passphrase

    def name(self) -> str:
        return self._name


class BackupDefinition(ABC):
    """Configuration definition containing credentials, tokens, ids"""

    _access: ServerAccess
    _type: str = ""
    _encryption: Encryption
    _collection_id: str
    _name: str
    _spec: dict
    _transport: TransportInterface

    def __init__(self, access: ServerAccess, _type: str, collection_id: str, encryption: Encryption,
                 name: str, spec: dict, transport: TransportInterface):
        self._access = access
        self._type = _type
        self._encryption = encryption
        self._collection_id = collection_id
        self._name = name
        self._spec = spec
        self._transport = transport

    @staticmethod
    def from_config(cls, config: dict, name: str):
        cls.validate_spec(config['spec'])

        return cls(
            access=config['meta']['access'],
            _type=config['meta']['type'],
            collection_id=config['meta']['collection_id'],
            encryption=config['meta']['encryption'],
            name=name,
            spec=config['spec'],
            transport=config['meta']['transport']
        )

    @staticmethod
    def get_spec_defaults() -> dict:
        """Specification default values"""

        return {}

    @classmethod
    def validate_spec(cls, spec: dict):
        spec_schema = cls.get_specification_schema()

        if spec_schema:
            try:
                validate(instance=spec, schema=spec_schema, format_checker=draft7_format_checker)
            except ValidationError as exc:
                raise SpecificationError(str(exc))

    @staticmethod
    @abstractmethod
    def get_specification_schema() -> dict:
        """A JSON-schema of 'spec' element of the configuration.
        'spec' is the specific configuration for given type of definition, for example MySQL will have a spec
        with instructions how to reach the database
        """
        pass

    @staticmethod
    def join_paths_quotes_into_string(paths: list):
        joined = ""

        for path in paths:
            joined += '"' + path + '" '

        return joined

    def access(self) -> ServerAccess:
        return self._access

    def encryption(self) -> Encryption:
        return self._encryption

    def get_type(self) -> str:
        return self._type

    def get_collection_id(self) -> str:
        return self._collection_id

    def transport(self) -> TransportInterface:
        return self._transport

    def get_sensitive_information(self) -> list:
        return []

    def name(self) -> str:
        return self._name

    def __repr__(self):
        return 'Definition<name=' + self._name + ',collection_id=' + str(self.get_collection_id()) + '>'

    @classmethod
    def get_example_configuration(cls) -> dict:
        schema = cls.get_specification_schema()

        if not schema or 'properties' not in schema:
            return {}

        return {
            'meta': {
                'type': cls.__module__,
                'access': 'my_backup_server',
                'encryption': 'enc_backup_1',
                'collection_id': '61792136-94d5-4670-9c69-950257467c56',
                'transport': 'local'
            },
            'spec': create_example_from_attributes(schema['properties'])
        }



import os
from io import BytesIO

from rkd.api.inputoutput import IO
from rkd.api.testing import BasicTestingCase
from bahub.exception import BufferingError
from bahub.inputoutput import StreamableBuffer


class TestStreamableBuffer(BasicTestingCase):
    def test_read_performs_pre_validation(self):
        buf = BytesIO(b'I\'ve never read Marx\'s Capital, but I\'ve got the marks of capital all over my body.')
        buf_obj = StreamableBuffer(
            io=IO(),
            read_callback=buf.read,
            close_callback=buf.close,
            is_success_callback=lambda: True,
            buffer=buf,
            has_exited_with_failure=lambda: True,  # read() should be checking this - we rely on this in this test
            description='Test stream via BytesIO',
            eof_callback=buf.closed,
            pre_validation_sleep=1
        )

        # we expect that after first bytes the stream pipe breaks
        # technically has_exited_with_failure() returns failure on first check, then exception is raised
        with self.assertRaises(BufferingError):
            buf_obj.read(8)

    def test_read_reads_whole_string_when_not_specifying_read_size(self):
        buf_obj = self._create_successful_example()

        self.assertEqual(
            b'I\'ve never read Marx\'s Capital, but I\'ve got the marks of capital all over my body.',
            buf_obj.read()
        )

    def test_read_reads_only_specified_amount_of_bytes(self):
        buf_obj = self._create_successful_example()

        self.assertEqual(
            b'I\'ve never',
            buf_obj.read(10)
        )

        self.assertEqual(
            b' read Marx\'s Capital',
            buf_obj.read(20)
        )

        self.assertEqual(
            b', but I\'ve got the marks of capital all over my body.',
            buf_obj.read()
        )

    def test_has_exited_with_failure_detects_that_parent_stream_exited(self):
        """
        Given Stream B is dependent on Stream B
        When I read from Stream B
        And Stream A fails
        Then Stream B also fails
        """

        # Parent stream
        parent_stream = BytesIO(
            b'4 Feb 1869, "Big" Bill Haywood was born. A miner from childhood, he co-founded the iww'
        )

        parent_stream_obj = StreamableBuffer(
            io=IO(),
            read_callback=parent_stream.read,
            close_callback=parent_stream.close,
            is_success_callback=lambda: True,
            buffer=parent_stream,
            has_exited_with_failure=lambda: True,
            description='Test stream via BytesIO',
            eof_callback=parent_stream.closed,
            pre_validation_sleep=0
        )

        # Current stream
        buf = BytesIO(b'I\'ve never read Marx\'s Capital, but I\'ve got the marks of capital all over my body.')
        buf_obj = StreamableBuffer(
            io=IO(),
            read_callback=buf.read,
            close_callback=buf.close,
            is_success_callback=lambda: True,
            buffer=buf,
            has_exited_with_failure=lambda: False,
            description='Test stream via BytesIO',
            eof_callback=buf.closed,
            pre_validation_sleep=0,
            parent=parent_stream_obj
        )

        # technically it should fail, because parent_stream_obj has has_exited_with_failure = True
        self.assertTrue(buf_obj.has_exited_with_failure())

    def test_factory_method_from_file(self):
        """
        Check that factory method properly opens file for reading
        """

        buf = StreamableBuffer.from_file(os.path.abspath(__file__), IO())
        text = buf.read()
        buf.close()

        self.assertIn(b'test_factory_method_from_file', text)
        self.assertTrue(buf.eof())

    @staticmethod
    def _create_successful_example() -> StreamableBuffer:
        """
        Example data provider
        :return: StreamableBuffer
        """

        buf = BytesIO(b'I\'ve never read Marx\'s Capital, but I\'ve got the marks of capital all over my body.')
        return StreamableBuffer(
            io=IO(),
            read_callback=buf.read,
            close_callback=buf.close,
            is_success_callback=lambda: True,
            buffer=buf,
            has_exited_with_failure=lambda: False,
            description='Test stream via BytesIO',
            eof_callback=buf.closed,
            pre_validation_sleep=0
        )



import os
from rkd.api.inputoutput import IO
from rkd.api.testing import BasicTestingCase
from rkd.yaml_parser import YamlFileLoader
from bahub.configurationfactory import ConfigurationFactory
from bahub.exception import ConfigurationFactoryException
from bahub.adapters.filesystem import Definition as FilesystemAdapterDefinition


TEST_DIR = os.path.dirname(os.path.realpath(__file__))
YAML_DIR = [TEST_DIR + '/env/config_factory_test', TEST_DIR + '/../bahub/internal']


class TestConfigurationFactory(BasicTestingCase):
    def test_parsing_envs_success_case(self):
        """
        See _process_env_variables()
        """

        with self.environment({'COLLECTION_ID': 'Rudolf Rocker'}):
            conf = ConfigurationFactory('bahub.test.conf.yaml', debug=True,
                                        parser=YamlFileLoader(YAML_DIR), io=IO())

            self.assertEqual(
                ['Father of anarchist syndicalism: Rudolf Rocker'],
                conf._process_env_variables(['Father of anarchist syndicalism: ${COLLECTION_ID}'])
            )

    def test_parsing_envs_raises_exception_when_variable_is_not_defined(self):
        with self.assertRaisesRegexp(ConfigurationFactoryException,
                                     'Following environment variables are not resolved: COLLECTION_ID'):
            conf = ConfigurationFactory('bahub.test.conf.yaml', debug=True,
                                        parser=YamlFileLoader(YAML_DIR), io=IO())

            conf._process_env_variables(['Father of anarchist syndicalism: ${COLLECTION_ID}'])

    def test_overall_parsing_definitions(self):
        """
        Functional test that relies strongly on test data provided in YAML_DIR
        """

        with self.environment({'COLLECTION_ID': 'Rudolf Rocker'}):
            conf = ConfigurationFactory('bahub.test.conf.yaml', debug=True,
                                        parser=YamlFileLoader(YAML_DIR), io=IO())

            definition = conf.get_definition('fs')

            with self.subTest('Check that backup type is correct'):
                self.assertEqual(FilesystemAdapterDefinition, type(definition))

            with self.subTest('Check connections between entities'):
                self.assertEqual(conf._transports['local'], definition.transport(), msg='Transport does not match')
                self.assertEqual(conf._encryption['strong'], definition.encryption())
                self.assertEqual(conf._accesses['secured'], definition.access())

            with self.subTest('Check basic information'):
                self.assertEqual('fs', definition.name())
                self.assertEqual({'paths': ['./']}, definition._spec)

    def test_functionally_get_all_sensitive_data(self):
        """
        Functional test that checks if credentials are listed from test data in YAML_DIR
        """

        with self.environment({'COLLECTION_ID': 'Rudolf Rocker'}):
            conf = ConfigurationFactory('bahub.test.conf.yaml', debug=True,
                                        parser=YamlFileLoader(YAML_DIR), io=IO())

            sensitive_data = conf.get_all_sensitive_data()

            self.assertIn('some-string-passphrase', sensitive_data)
            self.assertIn('eyJ0eXAiOiJKV1QiLCJhbGci', str(sensitive_data))

    def test_functionally_example_env_is_parsed(self):
        """
        Assert that "COLLECTION_ID" in example YAML file is parsed
        """

        with self.environment({'COLLECTION_ID': '1111-2222-3333-4444'}):
            conf = ConfigurationFactory('bahub.test.conf.yaml', debug=True,
                                        parser=YamlFileLoader(YAML_DIR), io=IO())

            self.assertEqual('1111-2222-3333-4444', conf.get_definition('fs').get_collection_id())



from rkd.api.testing import BasicTestingCase
from bahub.exception import ParsingException
from bahub.importing import Importing


class TestImporting(BasicTestingCase):
    def test_import_adapter_requires_adapter_class_to_be_defined(self):
        with self.assertRaisesRegexp(ParsingException,
                                     'Import "os" is invalid. Class or method "Adapter" not found in module'):
            Importing.import_adapter('os')

    def test_import_adapter_successfully_imports_adapter(self):
        adapter, definition = Importing.import_adapter('bahub.adapters.mysql')

        self.assertEqual('Adapter', adapter.__name__)
        self.assertEqual('Definition', definition.__name__)

    def test_import_transport_fails_when_transport_class_is_not_defined_in_module(self):
        with self.assertRaisesRegexp(ParsingException,
                                     'Import "sys" is invalid. Class or method "Transport" not found in module'):
            Importing.import_transport('sys')

    def test_import_transport_successfully_imports_transport(self):
        transport = Importing.import_transport('bahub.transports.sh')

        self.assertEqual('Transport', transport.__name__)



from abc import ABC as AbstractClass, abstractmethod
from rkd.api.inputoutput import IO
from ..exception import BackupRestoreError
from ..model import BackupDefinition
from ..transports.base import StreamableBuffer


class AdapterInterface(AbstractClass):
    @staticmethod
    @abstractmethod
    def create_definition(config: dict, name: str) -> BackupDefinition:
        """Creates a configuration object that will contain instructions how to eg. access the database, where to find
        files etc."""

        pass

    @abstractmethod
    def backup(self, definition: BackupDefinition) -> StreamableBuffer:
        pass

    @abstractmethod
    def restore(self, definition: BackupDefinition, in_buffer: StreamableBuffer, io: IO) -> None:
        """Restores a backup, returns a log as output"""

        pass

    @staticmethod
    def _read_from_restore_process(restore_process: StreamableBuffer, io: IO):
        io.print_separator()
        io.outln('Restore process output:')

        while not restore_process.eof():
            io.out(restore_process.read(1024 * 64).decode('utf-8'))

        io.print_separator()

        if restore_process.has_exited_with_failure():
            restore_process.close()

            raise BackupRestoreError.from_generic_restore_failure(restore_process)



from abc import abstractmethod
from typing import Union, List, Optional
from jsonschema import validate, draft7_format_checker, ValidationError
from rkd.api.inputoutput import IO
from ..exception import SpecificationError
from ..inputoutput import StreamableBuffer
from ..schema import create_example_from_attributes


class TransportInterface(object):
    """
    Transport
    =========

    Defines how to execute shell commands. It is not that simple - there are many possibilities:
        - Simply just execute on local shell
        - Execute inside running Docker container
        - Execute in Kubernetes container
        - Execute on remote machine
    """

    _spec: dict
    _io: IO

    def __init__(self, spec: dict, io: IO):
        self._spec = spec
        self._io = io

    def __enter__(self) -> 'TransportInterface':
        """
        Start using the transport. Here could be placed a code that will eg. spawn a docker container
        or notify the user - any action before usage of the Transport
        :return:
        """

        return self

    def __exit__(self, exc_type, exc_val, exc_t) -> None:
        """
        Finalize usage of the Transport - close what remained, clean up
        :return:
        """

        pass

    @staticmethod
    @abstractmethod
    def get_specification_schema() -> dict:
        """
        Should return a DICT with JSON Schema defining the "spec" of a transport in YAML

        :return:
        """

        return {}

    @classmethod
    def get_example_configuration(cls) -> dict:
        schema = cls.get_specification_schema()

        if not schema or 'properties' not in schema:
            return {}

        return {
            'type': cls.__module__,
            'spec': create_example_from_attributes(schema['properties'])
        }

    @classmethod
    def validate_spec(cls, spec: dict):
        spec_schema = cls.get_specification_schema()

        if spec_schema:
            try:
                validate(instance=spec, schema=spec_schema, format_checker=draft7_format_checker)
            except ValidationError as exc:
                raise SpecificationError('Linked transport "spec" section parsing error, ' + str(exc))

    @abstractmethod
    def execute(self, command: str):
        """
        Execute a command with passing through all output to console

        :param command:
        :return:
        """

        pass

    @abstractmethod
    def capture(self, command: Union[str, List[str]]) -> bytes:
        """
        Capture output

        :param command:
        :return:
        """

        pass

    @abstractmethod
    def buffered_execute(self, command: Union[str, List[str]],
                         stdin: Optional[StreamableBuffer] = None,
                         env: dict = None) -> StreamableBuffer:

        """
        Open a process with two pipes - IN & OUT for streaming

        :param command:
        :param stdin:
        :param env:
        :return:
        """

        pass

    def get_failure_details(self) -> str:
        """
        Optionally raise a specific exception with more details

        :return:
        """

        return ''

    def io(self) -> IO:
        return self._io



"""
Filesystem Adapter
==================

Packs files and directories into TAR.GZ packages
"""
import os

from rkd.api.inputoutput import IO
from .base import AdapterInterface
from ..model import BackupDefinition
from ..transports.base import StreamableBuffer


class Definition(BackupDefinition):
    """Configuration"""

    @staticmethod
    def get_spec_defaults() -> dict:
        return {}

    @staticmethod
    def get_specification_schema() -> dict:
        return {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "required": ["paths"],
            "properties": {
                "paths": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "example": ["/var/lib/jenkins"]
                }
            }
        }

    def get_backup_parameters(self):
        """
        Inside a package there could be multiple directories and files packaged from multiple paths
        Every path is quoted and checked before usage
        :return:
        """

        paths = []

        for path in self._spec.get('paths'):
            normalized = os.path.abspath(os.path.expanduser(path))

            if not normalized:
                continue

            paths.append('"{}"'.format(normalized))

        parameters = '-cz -f - ' + ' '.join(paths)

        return parameters

    def get_restore_parameters(self):
        """
        We unpack always at the root directory, because we allow to pack multiple paths
        :return:
        """

        return '-xzf - -C /'


class Adapter(AdapterInterface):
    """Defines how to make backup of files and directories"""

    def backup(self, definition: Definition) -> StreamableBuffer:
        """Pack files into a TAR.GZ and return as a output buffer"""

        backup_process = definition.transport().buffered_execute

        return backup_process('set -euo pipefail; tar %s | cat' % definition.get_backup_parameters())

    def restore(self, definition: Definition, in_buffer: StreamableBuffer, io: IO) -> None:
        """Unpack files from the TAR.GZ provided by in_buffer"""

        restore_process = definition.transport()\
            .buffered_execute('tar %s' % definition.get_restore_parameters(), stdin=in_buffer)

        self._read_from_restore_process(restore_process, io)

    @staticmethod
    def create_definition(config: dict, name: str) -> Definition:
        return BackupDefinition.from_config(Definition, config, name)



from argparse import ArgumentParser
from rkd.api.contract import ExecutionContext
from .base import BaseTask
from ..exception import InvalidResponseException
from ..inputoutput import StreamableBuffer


class UploaderTask(BaseTask):
    """Uploads a given file to a remote collection on the server
    """

    def get_name(self) -> str:
        return ':send'

    def get_group_name(self) -> str:
        return ':backup:raw'

    def configure_argparse(self, parser: ArgumentParser, with_definition: bool = True):
        super().configure_argparse(parser, with_definition=with_definition)
        parser.add_argument('--path', required=False,
                            help='File path from the disk. Defaults to empty and reading from stdin')

    def execute(self, context: ExecutionContext) -> bool:
        if not super().execute(context):
            return False

        definition_name = context.get_arg('definition')
        definition = self.config.get_definition(definition_name)
        source = self.read_source(context.get_arg('--path'))

        if not definition:
            self.io().error('Definition "{name}" is not valid'.format(name=definition_name))
            return False

        try:
            response = self.api.send_backup(
                collection_id=definition.get_collection_id(),
                access=definition.access(),
                attributes=definition.encryption().describe_as_attributes(),
                source=source
            )

            self.notifier.backup_was_uploaded(definition)

        except InvalidResponseException as response_exc:
            self.io().error(str(response_exc))
            self.io().error(response_exc.get_error())
            self.notifier.failed_to_upload_backup(definition, response_exc.get_error())

            return False

        if source:
            source.close()

        if response:
            self.io().outln(response.to_status_message())

        return True

    def read_source(self, path: str) -> StreamableBuffer:
        if not path:
            return StreamableBuffer.from_file('/dev/stdin', io=self._io)

        return StreamableBuffer.from_file(path, io=self._io)



#!/usr/bin/env python3

from setuptools import setup, find_packages

setup(
    setup_requires=['pbr'],
    pbr=True,
    package_dir={'': './'},
    packages=find_packages(where='./'),
    package_data={
        'bahub': [
            'internal/*',
            'internal/**/*',
            'internal/schema/*'
        ]
    }
)



"""
PostgreSQL Adapter
==================
"""
from typing import List

from rkd.api.inputoutput import IO
from ..model import BackupDefinition
from ..inputoutput import StreamableBuffer
from .base import AdapterInterface


class Definition(BackupDefinition):
    """Configuration"""

    @staticmethod
    def get_spec_defaults() -> dict:
        return {
            'port': 5432,
            'database': '',
            'gzip_args': '-3'
        }

    @staticmethod
    def get_specification_schema() -> dict:
        return {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "required": ["host", "user"],
            "properties": {
                "host": {
                    "type": "string"
                },
                "port": {
                    "type": "integer"
                },
                "database": {
                    "type": "string"
                },
                "user": {
                    "type": "string",
                },
                "password": {
                    "type": "string"
                },
                "gzip_args": {
                    "type": "string"
                }
            }
        }

    @classmethod
    def get_example_configuration(cls):
        return {
            'meta': {
                'type': 'bahub.adapters.postgres_dump',
                'access': 'my_backup_server',
                'encryption': 'enc_backup_db',
                'collection_id': '61792136-94d5-4670-9c69-950257467c56',
                'transport': 'local'
            },
            'spec': {
                'host': '127.0.0.1',
                'port': 5432,
                'database': 'gitea',
                'user': 'git_mdbDhSIfMFerfyAK',
                'password': 'boltcutter-goes-click-clack-KIVesvKc6dPIQ7scNsQsDg8mcc1x4SxQUVMjWPIq/VE=',
                'gzip_args': '-3'
            }
        }

    def get_sensitive_information(self) -> list:
        """
        Returns a list of keywords that needs to be stripped out from the console text
        :return:
        """

        return [
            self._spec['password']
        ]

    def _get_common_parameters(self) -> str:
        """
        Common commandline switches for mysql and mysqldump
        :return:
        """

        parameters = ' -h {host} -U {user} '.format(host=self._spec['host'], user=self._spec['user'])

        if self._spec.get('port'):
            parameters += ' -p {port} '.format(port=str(self._spec.get('port')))

        return parameters

    def is_dumping_all_databases(self):
        return not self._spec.get('database')

    def get_dump_command(self) -> str:
        """
        Commandline switches specific to mysqldump
        :return:
        """

        cmd = 'pg_dumpall' if self.is_dumping_all_databases() else 'pg_dump'
        cmd += self._get_common_parameters()
        cmd += ' --clean '

        if not self.is_dumping_all_databases():
            cmd += ' --set ON_ERROR_STOP=on '

        if self._spec.get('database'):
            cmd += ' {database} '.format(database=self._spec.get('database'))

        return cmd + ' | gzip ' + self.get_gzip_args()

    def get_restore_command(self) -> str:
        """
        Parameters for mysql command used in restore process
        :return:
        """

        parameters = 'gunzip | psql ' + self._get_common_parameters()

        if self._spec.get('database'):
            parameters += ' {database} '.format(database=self._spec.get('database'))

        return parameters

    def get_psql_command(self) -> str:
        return 'psql --tuples-only ' + self._get_common_parameters()

    def get_password(self) -> str:
        return self._spec.get('password')

    def get_gzip_args(self) -> str:
        return self._spec.get('gzip_args', '-3')


class Adapter(AdapterInterface):
    """
    PostgreSQL Backup Adapter
    =========================

    Understands how to backup & restore running PostgreSQL databases using pg_dump/pg_dumpall and psql basic tools

    Restore:
        1. All connected clients are kicked off
        2. Database connection limit is set to 0, so nobody can connect
        3. Database is restored
        4. The connection limits are restored to "unlimited"

    Backup:
        Uses just pg_dump/pg_dumpall.
    """

    def terminate_all_connections(self, definition: Definition):
        self.psql(
            sql='SELECT pg_terminate_backend(pid) FROM pg_stat_activity  WHERE pid <> pg_backend_pid()',
            definition=definition
        )

    def limit_connections_to_database(self, definition: Definition, db_name: str, limit: int):
        self.psql(
            sql='ALTER DATABASE ' + db_name + ' CONNECTION LIMIT ' + str(limit) + ';',
            definition=definition
        )

    def psql(self, sql: str, definition: Definition):
        return definition.transport()\
            .capture('echo "' + sql + '" | ' + definition.get_psql_command())

    def list_all_databases(self, definition: Definition) -> List[str]:
        output = self.psql('SELECT datname FROM pg_database WHERE datistemplate = false;', definition)

        return list(
            filter(
                lambda x: x,
                map(lambda part: part.strip().decode('utf-8'), output.split(b"\n"))
            )
        )

    def backup(self, definition: Definition) -> StreamableBuffer:
        """
        Starts a pg_dump/pg_dumpall process in buffering mode

        :param definition:
        :return:
        """

        return definition.transport()\
            .buffered_execute(
            definition.get_dump_command(),
            env={'PGPASSWORD': definition.get_password()}
        )

    def restore(self, definition: Definition, in_buffer: StreamableBuffer, io: IO) -> None:
        io.info('Terminating all database connections')
        self.terminate_all_connections(definition)
        dbs = self.list_all_databases(definition)

        for db in dbs:
            io.info('Setting connection limit=0 for database {}'.format(db))
            self.limit_connections_to_database(definition, db, limit=0)

        restore_process = definition.transport()\
            .buffered_execute(
                definition.get_restore_command(),
                stdin=in_buffer,
                env={'PGPASSWORD': definition.get_password()}
        )

        try:
            io.info('Performing restore using psql')
            self._read_from_restore_process(restore_process, io)
        finally:
            for db in dbs:
                io.info('Setting connection limit=unlimited for database {}'.format(db))
                self.limit_connections_to_database(definition, db, limit=-1)

    @staticmethod
    def create_definition(config: dict, name: str) -> Definition:
        return BackupDefinition.from_config(Definition, config, name)



"""
MySQL Adapter
=============
"""
from rkd.api.inputoutput import IO
from ..model import BackupDefinition
from ..inputoutput import StreamableBuffer
from .base import AdapterInterface


class Definition(BackupDefinition):
    """Configuration"""

    @staticmethod
    def get_spec_defaults() -> dict:
        return {
            'port': 3306,
            'database': '',
            'gzip_args': '-3'
        }

    @staticmethod
    def get_specification_schema() -> dict:
        return {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "required": ["host", "user"],
            "properties": {
                "host": {
                    "type": "string"
                },
                "port": {
                    "type": "integer"
                },
                "database": {
                    "type": "string"
                },
                "user": {
                    "type": "string",
                },
                "password": {
                    "type": "string"
                },
                "gzip_args": {
                    "type": "string"
                }
            }
        }

    @classmethod
    def get_example_configuration(cls):
        return {
            'meta': {
                'type': 'bahub.adapters.mysql',
                'access': 'my_backup_server',
                'encryption': 'enc_backup_db',
                'collection_id': '61792136-94d5-4670-9c69-950257467c56',
                'transport': 'local'
            },
            'spec': {
                'host': '127.0.0.1',
                'port': 3306,
                'database': 'gitea',
                'user': 'git_mdbDhSIfMFerfyAK',
                'password': 'boltcutter-goes-click-clack-KIVesvKc6dPIQ7scNsQsDg8mcc1x4SxQUVMjWPIq/VE=',
                'gzip_args': '-3'
            }
        }

    def get_sensitive_information(self) -> list:
        """
        Returns a list of keywords that needs to be stripped out from the console text
        :return:
        """

        return [
            self._spec['password']
        ]

    def _get_common_parameters(self) -> str:
        """
        Common commandline switches for mysql and mysqldump
        :return:
        """

        parameters = '-h {host} -u {user} '.format(host=self._spec['host'], user=self._spec['user'])

        if self._spec.get('password'):
            parameters += ' -p"{password}" '.format(password=self._spec.get('password').replace('$', '\$'))

        if self._spec.get('port'):
            parameters += ' -P {port} '.format(port=str(self._spec.get('port')))

        return parameters

    def get_dump_parameters(self) -> str:
        """
        Commandline switches specific to mysqldump
        :return:
        """

        parameters = self._get_common_parameters()
        parameters += ' --skip-lock-tables --add-drop-table --add-drop-database --add-drop-trigger '

        if self._spec.get('database'):
            parameters += ' {database} '.format(database=self._spec.get('database'))
        else:
            parameters += ' --all-databases '

        return parameters

    def get_restore_parameters(self) -> str:
        """
        Parameters for mysql command used in restore process
        :return:
        """

        parameters = self._get_common_parameters()

        if self._spec.get('database'):
            parameters += ' {database} '.format(database=self._spec.get('database'))

        return parameters

    def get_gzip_args(self) -> str:
        return self._spec.get('gzip_args', '-3')


class Adapter(AdapterInterface):
    """Contains a logic specific to MySQL - how to backup, and how to restore"""

    def backup(self, definition: Definition) -> StreamableBuffer:
        backup_process = definition.transport().buffered_execute

        return backup_process('mysqldump {} | gzip {}'.format(
            definition.get_dump_parameters(),
            definition.get_gzip_args()
        ))

    def restore(self, definition: Definition, in_buffer: StreamableBuffer, io: IO) -> None:
        restore_process = definition.transport().buffered_execute('gunzip | mysql %s'
                                                                  % definition.get_restore_parameters(),
                                                                  stdin=in_buffer)

        self._read_from_restore_process(restore_process, io)

    @staticmethod
    def create_definition(config: dict, name: str) -> Definition:
        return BackupDefinition.from_config(Definition, config, name)



from rkd.api.inputoutput import IO
from rkd.api.testing import BasicTestingCase

from bahub.exception import DockerContainerError
from bahub.transports.docker import Transport


class TestDockerTransport(BasicTestingCase):
    """
    Functional test - requires docker daemon and docker client tools
    """

    def test_captures_output(self):
        self.assertIn(b'usr', self._create_example_transport().capture('ls -la /').strip())

    def test_buffers_output(self):
        buffer = self._create_example_transport().buffered_execute('ls -la /')
        buffer._pre_validation_sleep = 0  # hack to only speed up tests, normally it is a private method

        self.assertIn(b'usr', buffer.read().strip())
        buffer.close()

    def test_container_is_validated(self):
        """
        Expects that all methods that executes commands will trigger container validation
        """

        invalid_container_transport = Transport(
            spec={
                'container': 'container_name_here_is_invalid',
                'shell': '/bin/sh'
            },
            io=IO()
        )

        with self.subTest('capture()'):
            with self.assertRaises(DockerContainerError):
                invalid_container_transport.capture('ls')

        with self.subTest('buffered_execute()'):
            with self.assertRaises(DockerContainerError):
                invalid_container_transport.buffered_execute('ls')

        with self.subTest('execute()'):
            with self.assertRaises(DockerContainerError):
                invalid_container_transport.execute('ls')

    @staticmethod
    def _create_example_transport() -> Transport:
        return Transport(
            spec={
                'container': 's3pb_db_mysql_1',
                'shell': '/bin/sh'
            },
            io=IO()
        )
